# -*- encoding: utf-8 -*-
# merge_args test suite
# Copyright © 2018, Chris Warrick.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the author of this software nor the names of
#    contributors to this software may be used to endorse or promote
#    products derived from this software without specific prior written
#    consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



# -*- encoding: utf-8 -*-
# merge_args v0.1.0
# Merge signatures of two functions with Advanced Hackery.
# Copyright © 2018, Chris Warrick.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the author of this software nor the names of
#    contributors to this software may be used to endorse or promote
#    products derived from this software without specific prior written
#    consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Merge signatures of two functions with Advanced Hackery. Useful for wrappers.

Usage: @merge_args(old_function)
"""

import inspect
import itertools
import types
import functools

__version__ = '0.1.0'
__all__ = ('merge_args',)


def _blank():  # pragma: no cover
    pass


def _merge(source, dest):
    """Merge the signatures of ``source`` and ``dest``.

    ``dest`` args go before ``source`` args in all three categories
    (positional, keyword-maybe, keyword-only).
    """
    source_spec = inspect.getfullargspec(source)
    dest_spec = inspect.getfullargspec(dest)

    if source_spec.varargs or source_spec.varkw:
        raise ValueError("The source function may not take variable arguments.")

    source_all = source_spec.args
    dest_all = dest_spec.args

    if source_spec.defaults:
        source_pos = source_all[:-len(source_spec.defaults)]
        source_kw = source_all[-len(source_spec.defaults):]
    else:
        source_pos = source_all
        source_kw = []

    if dest_spec.defaults:
        dest_pos = dest_all[:-len(dest_spec.defaults)]
        dest_kw = dest_all[-len(dest_spec.defaults):]
    else:
        dest_pos = dest_all
        dest_kw = []

    args_merged = dest_pos

    for a in itertools.chain(source_pos, dest_kw, source_kw):
        if a not in args_merged:
            args_merged.append(a)

    kwonlyargs_merged = dest_spec.kwonlyargs
    for a in source_spec.kwonlyargs:
        if a not in kwonlyargs_merged:
            kwonlyargs_merged.append(a)

    args_all = tuple(args_merged + kwonlyargs_merged)

    passer_code = types.CodeType(len(args_merged), len(kwonlyargs_merged),
                                 _blank.__code__.co_nlocals,
                                 _blank.__code__.co_stacksize,
                                 source.__code__.co_flags,
                                 _blank.__code__.co_code, (), (),
                                 args_all, dest.__code__.co_filename,
                                 dest.__code__.co_name,
                                 dest.__code__.co_firstlineno,
                                 dest.__code__.co_lnotab)
    passer = types.FunctionType(passer_code, globals())
    dest.__wrapped__ = passer

    # defaults, annotations
    defaults = dest.__defaults__
    if defaults is None:
        defaults = []
    else:
        defaults = list(defaults)
    if source.__defaults__ is not None:
        defaults.extend(source.__defaults__)

    # ensure we take destination’s return annotation
    has_dest_ret = 'return' in dest.__annotations__
    if has_dest_ret:
        dest_ret = dest.__annotations__['return']

    for v in ('__kwdefaults__', '__annotations__'):
        out = getattr(source, v)
        if out is None:
            out = {}
        if getattr(dest, v) is not None:
            out = out.copy()
            out.update(getattr(dest, v))
            setattr(passer, v, out)

    if has_dest_ret:
        passer.__annotations__['return'] = dest_ret
    dest.__annotations__ = passer.__annotations__

    passer.__defaults__ = tuple(defaults)
    if not dest.__doc__:
        dest.__doc__ = source.__doc__
    return dest


def merge_args(source):
    """Merge the signatures of two functions."""
    return functools.partial(_merge, source)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# merge_args test suite
# Copyright © 2018, Chris Warrick.
# See /LICENSE for licensing information.

import pytest
import merge_args
from inspect import signature


def test_orig():

    def old(foo, bar):
        """This is old's docstring."""
        return foo + bar

    def new(prefix, foo, *args, **kwargs):
        return old(prefix + foo, *args, **kwargs)

    assert str(signature(new)) == '(prefix, foo, *args, **kwargs)'

    new = merge_args._merge(old, new)
    assert str(signature(new)) == '(prefix, foo, bar)'
    assert new.__doc__ == old.__doc__
    assert new.__doc__
    assert old('a', 'b') == 'ab'
    assert new('!', 'a', 'b') == '!ab'


def test_with_kwargs():

    def src(a, b, c=1): pass

    @merge_args.merge_args(src)
    def dest(d, e=2, *args, **kwargs): pass

    assert str(signature(dest)) == '(d, a, b, e=2, c=1)'


def test_kwonlyargs():

    def src(q, w, e=1, *, r=2): pass

    @merge_args.merge_args(src)
    def dest(*, t=3, **kwargs): pass

    assert str(signature(dest)) == '(q, w, e=1, *, t=3, r=2)'

    def src2(*, a=1): pass

    @merge_args.merge_args(src2)
    def dest2(*, a=2, b=3): pass

    assert str(signature(dest2)) == '(*, a=2, b=3)'


def test_var_fails():

    def src(*args, **kwargs): pass

    def dest(*args, **kwargs): pass

    with pytest.raises(ValueError):
        merge_args._merge(src, dest)


def test_no_docstring_override():

    def src(foo):
        """src docs"""

    @merge_args.merge_args(src)
    def dest(bar):
        """dest docs"""

    assert str(signature(dest)) == '(bar, foo)'
    assert src.__doc__ == "src docs"
    assert dest.__doc__ == "dest docs"


def test_annotations():

    def src(s: int) -> str:
        pass

    @merge_args.merge_args(src)
    def dest(d: float) -> list:
        pass

    assert str(signature(dest)) == '(d: float, s: int) -> list'



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import io
from setuptools import setup


setup(name='merge_args',
      version='0.1.0',
      description='Merge signatures of two functions with Advanced Hackery.',
      keywords='merge_args',
      author='Chris Warrick',
      author_email='chris@chriswarrick.com',
      url='https://github.com/Kwpolska/merge_args',
      license='3-clause BSD',
      long_description=io.open(
          './README.rst', 'r', encoding='utf-8').read(),
      platforms='any',
      zip_safe=False,
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=['Development Status :: 1 - Planning',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   ],
      py_modules=['merge_args'],
      )
